/* Esco-GDM: esco-gdm-config.h
 * 
 * Copyright (C) 2004 James M. Cape <jcape@ignore-your.tv>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include "esco-gdm-config.h"
#include "ve-config.h"
#include "ve-misc.h"


typedef struct _ConfigData
{
  guint ref;

  gchar *global_face_dir;
  gssize max_face_size;
  GHashTable *excluded_users;
  gulong min_uid;
  EscoRelaxPermsFlags relax_perms:2;
  gboolean allow_root:1;
}
ConfigData;


static ConfigData *config = NULL;


/* ************************ *
 *  Housekeeping Functions  *
 * ************************ */

void
esco_gdm_config_init (void)
{
  const gchar *const config_files[] = {
    SYSCONFDIR "/gdm/gdm.conf",
    SYSCONFDIR "/X11/gdm/gdm.conf",
    "/etc/gdm/gdm.conf",
    "/etc/X11/gdm/gdm.conf"
  };
  VeConfig *cfg;
  guint i;

  if (config != NULL)
    {
      config->ref++;
      return;
    }

  config = g_new0 (ConfigData, 1);
  config->ref = 1;
  config->excluded_users = NULL;
  config->min_uid = 0;
  config->global_face_dir = NULL;
  config->max_face_size = -1;
  config->allow_root = TRUE;

  cfg = NULL;

  for (i = 0; i < G_N_ELEMENTS (config_files) && cfg == NULL; i++)
    {
      if (g_file_test (config_files[i], G_FILE_TEST_EXISTS))
	cfg = ve_config_new (config_files[i]);
    }

  if (cfg != NULL)
    {
      gchar *tmp;
      gint tmp_int;

      tmp = ve_config_get_string (cfg, "greeter/Exclude");

      if (tmp != NULL)
	{
	  gchar **users;

	  config->excluded_users = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	  users = g_strsplit (tmp, ",", -1);

	  for (i = 0; users[i] != NULL; i++)
	    g_hash_table_insert (config->excluded_users, users[i], GUINT_TO_POINTER (!NULL));

	  g_free (users);
	  g_free (tmp);
	}

      tmp_int = ve_config_get_int (cfg, "greeter/MinimalUID");
      if (tmp_int <= 0)
	config->min_uid = 0;
      else
	config->min_uid = (guint) tmp_int;

      config->global_face_dir = ve_config_get_string (cfg, "greeter/GlobalFaceDir");
      
      tmp_int = ve_config_get_int (cfg, "security/UserMaxFile");
      if (tmp_int <= 0)
	config->max_face_size = G_MAXINT;
      else
	config->max_face_size = tmp_int;

      config->allow_root = ve_config_get_bool (cfg, "security/AllowRoot");

      tmp_int = ve_config_get_int (cfg, "security/RelaxPermissions");

      switch (tmp_int)
	{
	case 1:
	  config->relax_perms = ESCO_RELAX_PERMS_GROUP;
	  break;
	case 2:
	  config->relax_perms = ESCO_RELAX_PERMS_ALL;
	  break;

	default:
	  config->relax_perms = ESCO_RELAX_PERMS_NONE;
	  break;
	}

      ve_config_destroy (cfg);
    }

  if (config->global_face_dir == NULL)
    config->global_face_dir = g_build_filename (DATADIR, "faces", NULL);

  if (config->max_face_size <= 0)
    config->max_face_size = G_MAXINT;
}


void
esco_gdm_config_shutdown (void)
{
  if (config == NULL)
    return;

  config->ref--;

  if (config->ref > 0)
    return;

  if (config->excluded_users != NULL)
    g_hash_table_destroy (config->excluded_users);

  g_free (config->global_face_dir);
  g_free (config);
  config = NULL;
}


EscoRelaxPermsFlags
esco_gdm_config_get_relax_perms (void)
{
  EscoRelaxPermsFlags retval;

  esco_gdm_config_init ();
  retval = config->relax_perms;
  esco_gdm_config_shutdown ();

  return retval;
}


static void
listify_table_values (gpointer key, gpointer value, GList **list)
{
  *list = g_list_prepend (*list, g_strdup (value));
}


GList *
esco_gdm_config_get_excluded_users (void)
{
  GList *retval = NULL;

  esco_gdm_config_init ();
  g_hash_table_foreach (config->excluded_users, (GHFunc) listify_table_values, &retval);
  esco_gdm_config_shutdown ();

  return retval;
}


gboolean
esco_gdm_config_is_user_excluded (const gchar *username)
{
  gboolean retval;

  g_return_val_if_fail (!ve_string_empty (username), FALSE);

  esco_gdm_config_init ();
  retval = (g_hash_table_lookup (config->excluded_users, username) != NULL);
  esco_gdm_config_shutdown ();

  return retval;
}


gchar *
esco_gdm_config_get_global_face_dir (void)
{
  gchar *retval;

  esco_gdm_config_init ();
  retval = g_strdup (config->global_face_dir);
  esco_gdm_config_shutdown ();

  return retval;
}


gssize
esco_gdm_config_get_max_user_size (void)
{
  gssize retval;

  esco_gdm_config_init ();
  retval = config->max_face_size;
  esco_gdm_config_shutdown ();

  return retval;
}


gboolean
esco_gdm_config_get_allow_root (void)
{
  gboolean retval;

  esco_gdm_config_init ();
  retval = config->allow_root;
  esco_gdm_config_shutdown ();

  return retval;
}


gulong
esco_gdm_config_get_minimal_uid (void)
{
  gulong retval;

  esco_gdm_config_init ();
  retval = config->min_uid;
  esco_gdm_config_shutdown ();

  return retval;
}
