/* Esco-GDM: esco-screen-manager.c
 * 
 * Copyright (C) 2004 James M. Cape <jcape@ignore-your.tv>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <stdio.h>

#include <glib/gi18n.h>

#include <X11/Xauth.h>

#include "gdmcomm.h"
#include "esco-screen-manager.h"


/* ******** *
 *  Macros  *
 * ******** */

#define DISPLAY(strv)	((strv)[0])
#define USERNAME(strv)	((strv)[1])
#define VT(strv)	((strv)[2])


/* ************** *
 *  Enumerations  *
 * ************** */

enum
{
  PROP_0,
  SCREENS
};

enum
{
  NEW_SCREEN,
  SCREEN_CHANGED,
  SCREEN_CLOSED,

  LAST_SIGNAL
};


/* ************ *
 *  Structures  *
 * ************ */

struct _EscoScreenManager
{
  GObject parent;

  GList *screens;
  gchar *vtcache;
};

struct _EscoScreenManagerClass
{
  GObjectClass parent_class;

  void (*new_screen)      (EscoScreenManager *mgr, EscoScreen *screen);
  void (*screen_changed)  (EscoScreenManager *mgr, EscoScreen *screen);
  void (*screen_closed)   (EscoScreenManager *mgr, EscoScreen *screen);
};


/* ****************** *
 *  Global Variables  *
 * ****************** */

static gulong esco_screen_manager_signals[LAST_SIGNAL] = { 0 };


/* ******************* *
 *  GType Declaration  *
 * ******************* */

G_DEFINE_TYPE (EscoScreenManager, esco_screen_manager, G_TYPE_OBJECT);


/* ******************* *
 *  Utility Functions  *
 * ******************* */

static inline EscoScreen *
find_screen_by_vt (GList *screens, guint vt)
{
  while (screens != NULL)
    {
      if (ESCO_SCREEN (screens->data)->vt == vt)
	return screens->data;

      screens = screens->next;
    }

  return NULL;
}


/* ******************* *
 *  GObject Functions  *
 * ******************* */

static void
esco_screen_manager_get_property (GObject * object, guint id, GValue * value, GParamSpec * pspec)
{
  EscoScreenManager *mgr = ESCO_SCREEN_MANAGER (object);

  switch (id)
    {
    case SCREENS:
      {
	GList *list;
	GValueArray *ar;

	ar = g_value_array_new (g_list_length (mgr->screens));

	for (list = mgr->screens; list != NULL; list = list->next)
	  {
	    GValue tmp_value = { 0 };

	    g_value_init (&tmp_value, ESCO_TYPE_SCREEN);
	    g_value_set_boxed (&tmp_value, list->data);
	    g_value_array_prepend (ar, &tmp_value);
	    g_value_unset (&tmp_value);
	  }

	g_value_take_boxed (value, ar);
      }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}

static void
esco_screen_manager_finalize (GObject * object)
{
  EscoScreenManager *mgr = ESCO_SCREEN_MANAGER (object);

  g_free (mgr->vtcache);
  g_list_foreach (mgr->screens, (GFunc) esco_screen_unref, NULL);
  g_list_free (mgr->screens);

  if (G_OBJECT_CLASS (esco_screen_manager_parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS (esco_screen_manager_parent_class)->finalize) (object);
}


/* ***************** *
 *  GType Functions  *
 * ***************** */

static void
esco_screen_manager_class_init (EscoScreenManagerClass * class)
{
  GObjectClass *object_class;

  object_class = G_OBJECT_CLASS (class);

  object_class->get_property = esco_screen_manager_get_property;
  object_class->finalize = esco_screen_manager_finalize;

  g_object_class_install_property (object_class, SCREENS,
				   g_param_spec_value_array ("screens", "Screens",
							     "A value array containing the open "
							     "screens",
							     g_param_spec_boxed ("screen", "Screen",
										 "A single screen",
										 ESCO_TYPE_SCREEN,
										 G_PARAM_READABLE),
							     G_PARAM_READABLE));

  esco_screen_manager_signals[NEW_SCREEN] =
    g_signal_new ("new-screen",
		  G_TYPE_FROM_CLASS (class),
		  (G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION),
		  G_STRUCT_OFFSET (EscoScreenManagerClass, new_screen),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__BOXED,
		  G_TYPE_NONE, 1, ESCO_TYPE_SCREEN);
  esco_screen_manager_signals[SCREEN_CHANGED] =
    g_signal_new ("screen-changed",
		  G_TYPE_FROM_CLASS (class),
		  (G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION),
		  G_STRUCT_OFFSET (EscoScreenManagerClass, screen_changed),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__BOXED,
		  G_TYPE_NONE, 1, ESCO_TYPE_SCREEN);
  esco_screen_manager_signals[SCREEN_CLOSED] =
    g_signal_new ("screen-closed",
		  G_TYPE_FROM_CLASS (class),
		  (G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION),
		  G_STRUCT_OFFSET (EscoScreenManagerClass, screen_changed),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__BOXED,
		  G_TYPE_NONE, 1, ESCO_TYPE_SCREEN);
}


static void
esco_screen_manager_init (EscoScreenManager * mgr)
{
  mgr->screens = NULL;
}


/* ************************************************************************** *
 *  Public API                                                                *
 * ************************************************************************** */

EscoScreenManager *
esco_screen_manager_new (void)
{
  return g_object_new (ESCO_TYPE_SCREEN_MANAGER, NULL);
}


EscoScreenManager *
esco_screen_manager_get_default (void)
{
  static EscoScreenManager *mgr = NULL;

  if (mgr == NULL)
    mgr = esco_screen_manager_new ();

  return mgr;
}


gboolean
esco_screen_manager_update (EscoScreenManager *mgr, GError **error)
{
  const gchar *auth_cookie;
  gchar *result;

  auth_cookie = gdmcomm_get_auth_cookie ();
  if (auth_cookie == NULL)
    {
      if (error != NULL)
	{
	  *error = g_error_new (ESCO_GDM_ERROR, ESCO_GDM_ERROR_AUTH_NOT_AVAILABLE,
				_("The authentication file, %s, could not be read."),
				XauFileName ());
	}

      return FALSE;
    }

  result = gdmcomm_call_gdm (GDM_SUP_CONSOLE_SERVERS, auth_cookie, "2.2.4.0", 1);

  if (result == NULL || strncmp (result, "OK", 2) != 0)
    {
      if (error != NULL)
	{
	  gint code;

	  if (result != NULL)
	    sscanf (result + 4, "%d", &code);
	  else
	    code = ESCO_GDM_ERROR_GDM_NOT_RUNNING;

	  *error = g_error_new_literal (ESCO_GDM_ERROR, code,
					gdmcomm_get_error_message (result, FALSE));
	}

      g_free (result);
      return FALSE;
    }

  if (mgr->vtcache == NULL || strcmp (result + 3, mgr->vtcache) != 0)
    {
      GList *current_screens;
      gchar **screens_data;
      guint i;

      current_screens = NULL;

      g_free (mgr->vtcache);
      mgr->vtcache = g_strdup (result + 3);

      screens_data = g_strsplit (mgr->vtcache, ";", -1);
      for (i = 0; screens_data[i] != NULL; i++)
	{
	  EscoScreen *screen;
	  gchar **screen_data;
	  guint vt;

	  screen_data = g_strsplit (screens_data[i], ",", 3);

	  /* This is an invalid result! */
	  if (DISPLAY (screen_data) == NULL
	      || USERNAME (screen_data) == NULL
	      || VT (screen_data) == NULL
	      || sscanf (VT (screen_data), "%u", &vt) < 1)
	    {
	      g_strfreev (screen_data);
	      continue;
	    }

	  g_free (VT (screen_data));

	  screen = find_screen_by_vt (mgr->screens, vt);

	  if (screen == NULL)
	    {
	      screen = g_new0 (EscoScreen, 1);
	      screen->g_class.g_type = ESCO_TYPE_SCREEN;
	      screen->ref = 1;

	      screen->vt = vt;

	      screen->display = DISPLAY (screen_data);

	      if (USERNAME (screen_data)[0] == '\0')
		{
		  screen->username = NULL;
		  g_free (USERNAME (screen_data));
		}
	      else
		{
		  screen->username = USERNAME (screen_data);
		}

	      g_free (screen_data);

	      g_signal_emit (mgr, esco_screen_manager_signals[NEW_SCREEN], 0, screen);
	    }
	  else
	    {
	      gboolean dirty;

	      esco_screen_ref (screen);

	      if (screen->username == NULL)
		{
		  if (USERNAME (screen_data)[0] == '\0')
		    {
		      g_free (USERNAME (screen_data));
		      dirty = FALSE;
		    }
		  else
		    {
		      screen->username = USERNAME (screen_data);
		      dirty = TRUE;
		    }
		}
	      else
		{
		  if (strcmp (screen->username, USERNAME (screen_data)) != 0)
		    {
		      g_free (screen->username);

		      if (USERNAME (screen_data)[0] == '\0')
			{
			  screen->username = NULL;
			  g_free (USERNAME (screen_data));
			}
		      else
			{			  
			  screen->username = USERNAME (screen_data);
			}

		      dirty = TRUE;
		    }
		  else
		    {
		      dirty = FALSE;
		    }
		}

	      if (dirty)
		g_signal_emit (mgr, esco_screen_manager_signals[SCREEN_CHANGED], 0, screen);
	    }

	  current_screens = g_list_prepend (current_screens, screen);
	}

      while (mgr->screens != NULL)
	{
	  if (g_list_find (current_screens, mgr->screens->data) == NULL)
	    {
	      g_signal_emit (mgr, esco_screen_manager_signals[SCREEN_CLOSED], 0,
			     mgr->screens->data);
	    }

	  esco_screen_unref (mgr->screens->data);
	  mgr->screens = g_list_remove_link (mgr->screens, mgr->screens);
	}

      mgr->screens = current_screens;
    }

  g_free (result);
  return TRUE;
}



GList *
esco_screen_manager_list_screens (EscoScreenManager *mgr)
{
  GList *list, *retval;

  g_return_val_if_fail (ESCO_IS_SCREEN_MANAGER (mgr), NULL);

  retval = NULL;
  for (list = mgr->screens; list != NULL; list = list->next)
    {
      retval = g_list_prepend (retval, esco_screen_ref (list->data));
    }

  return retval;
}


GList *
esco_screen_manager_list_screens_for_user (EscoScreenManager *mgr, const gchar *username)
{
  GList *list, *retval;

  g_return_val_if_fail (ESCO_IS_SCREEN_MANAGER (mgr), NULL);
  g_return_val_if_fail (username != NULL && username[0] != '\0', NULL);

  retval = NULL;
  for (list = mgr->screens; list != NULL; list = list->next)
    {
      if ((username == NULL &&
	   ESCO_SCREEN (list->data)->username == NULL) ||
	  (ESCO_SCREEN (list->data)->username != NULL &&
	   strcmp (ESCO_SCREEN (list->data)->username, username) == 0))
	{
	  retval = g_list_prepend (retval, esco_screen_ref (list->data));
	}
    }

  return retval;
}


EscoScreen *
esco_screen_manager_get_this_screen (EscoScreenManager *mgr, GError **error)
{
  const gchar *auth_cookie;
  gchar *result;
  EscoScreen *retval;

  g_return_val_if_fail (ESCO_IS_SCREEN_MANAGER (mgr), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  retval = NULL;
  auth_cookie = gdmcomm_get_auth_cookie ();

  if (auth_cookie == NULL)
    {
      if (error != NULL)
	{
	  *error = g_error_new (ESCO_GDM_ERROR, ESCO_GDM_ERROR_AUTH_NOT_AVAILABLE,
				_("The authentication file, %s, could not be read."),
				XauFileName ());
	}

      return NULL;
    }

  result = gdmcomm_call_gdm (GDM_SUP_QUERY_VT, auth_cookie, "2.2.4.0", 1);

  if (result == NULL || strncmp (result, "OK", 2) != 0)
    {
      if (error != NULL)
	{
	  gint code;

	  if (result != NULL)
	    sscanf (result + 4, "%d", &code);
	  else
	    code = ESCO_GDM_ERROR_GDM_NOT_RUNNING;

	  *error = g_error_new_literal (ESCO_GDM_ERROR, code,
					gdmcomm_get_error_message (result, FALSE));
	}

      retval = NULL;
    }
  else
    {
      guint vt;
      
      if (sscanf (result + 3, "%u", &vt) < 1)
	retval = FALSE;
      else
	retval = find_screen_by_vt (mgr->screens, vt);
    }

  g_free (result);

  return (retval != NULL ? esco_screen_ref (retval) : retval);
}


EscoScreen *
esco_screen_manager_get_screen (EscoScreenManager *mgr, guint vt)
{
  EscoScreen *retval;

  g_return_val_if_fail (ESCO_IS_SCREEN_MANAGER (mgr), NULL);

  retval = find_screen_by_vt (mgr->screens, vt);

  return (retval != NULL ? esco_screen_ref (retval) : retval);
}


gboolean
esco_screen_manager_set_screen (EscoScreenManager *mgr, EscoScreen *screen, GError **error)
{
  const gchar *auth_cookie;
  gchar *command, *result;
  gboolean retval;

  g_return_val_if_fail (ESCO_IS_SCREEN_MANAGER (mgr), FALSE);
  g_return_val_if_fail (ESCO_IS_SCREEN (screen), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  auth_cookie = gdmcomm_get_auth_cookie ();

  if (auth_cookie == NULL)
    {
      if (error != NULL)
	{
	  *error = g_error_new (ESCO_GDM_ERROR, ESCO_GDM_ERROR_AUTH_NOT_AVAILABLE,
				_("The authentication file, %s, could not be read."),
				XauFileName ());
	}

      return FALSE;
    }

  command = g_strdup_printf ("%s %u", GDM_SUP_SET_VT, screen->vt);

  result = gdmcomm_call_gdm (command, auth_cookie, "2.5.90.0", 1);
  g_free (command);

  if (result == NULL || strncmp (result, "OK", 2) != 0)
    {
      if (error != NULL)
	{
	  gint code;

	  if (result != NULL)
	    sscanf (result + 4, "%d", &code);
	  else
	    code = ESCO_GDM_ERROR_GDM_NOT_RUNNING;

	  *error = g_error_new_literal (ESCO_GDM_ERROR, code,
					gdmcomm_get_error_message (result, FALSE));
	}

      retval = FALSE;
    }
  else
    {
      retval = TRUE;
    }

  g_free (result);
  
  return retval;
}


gboolean
esco_screen_manager_new_screen (EscoScreenManager *mgr, GError **error)
{
  GList *list;
  const gchar *auth_cookie;
  gchar *command, *result;
  gboolean retval;
  guint vt;

  g_return_val_if_fail (ESCO_IS_SCREEN_MANAGER (mgr), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  auth_cookie = gdmcomm_get_auth_cookie ();
  if (auth_cookie == NULL)
    {
      if (error != NULL)
	{
	  *error = g_error_new (ESCO_GDM_ERROR, ESCO_GDM_ERROR_AUTH_NOT_AVAILABLE,
				_("The authentication file, %s, could not be read."),
				XauFileName ());
	}

      return FALSE;
    }

  /* Try and find an existing empty vt to switch to first */
  vt = 0;
  for (list = mgr->screens; list != NULL && vt == 0; list = list->next)
    {
      if (ESCO_SCREEN (list->data)->username == NULL)
	vt = ESCO_SCREEN (list->data)->vt;
    }

  if (vt != 0)
    command = g_strdup_printf ("%s %u", GDM_SUP_SET_VT, vt);
  else
    command = g_strdup (GDM_SUP_FLEXI_XSERVER);

  result = gdmcomm_call_gdm (command, auth_cookie, "2.5.90.0", 1);
  g_free (command);

  if (result == NULL || strncmp (result, "OK", 2) != 0)
    {
      if (error != NULL)
	{
	  gint code;

	  if (result != NULL)
	    sscanf (result + 4, "%d", &code);
	  else
	    code = ESCO_GDM_ERROR_GDM_NOT_RUNNING;

	  *error = g_error_new_literal (ESCO_GDM_ERROR, code,
					gdmcomm_get_error_message (result, FALSE));
	}

      retval = FALSE;
    }
  else
    {
      retval = TRUE;
    }

  g_free (result);
  
  return retval;
}


GType
esco_screen_get_type (void)
{
  static GType type = G_TYPE_INVALID;

  if (type == G_TYPE_INVALID)
    {
      type = g_boxed_type_register_static ("EscoScreen", (GBoxedCopyFunc) esco_screen_ref,
					   (GBoxedFreeFunc) esco_screen_unref);
    }

  return type;
}


EscoScreen *
esco_screen_ref (EscoScreen *screen)
{
  g_return_val_if_fail (screen == NULL || ESCO_IS_SCREEN (screen), NULL);

  if (screen == NULL)
    return NULL;

  screen->ref++;

  return screen;
}


void
esco_screen_unref (EscoScreen *screen)
{
  g_return_if_fail (screen == NULL || ESCO_IS_SCREEN (screen));

  if (screen == NULL)
    return;

  screen->ref--;

  if (screen->ref > 0)
    return;

  g_free (screen->display);
  g_free (screen->username);
  g_free (screen);
}


GQuark
esco_gdm_error_get_quark (void)
{
  static GQuark quark = 0;

  if (quark == 0)
    quark = g_quark_from_static_string ("esco-gdm-error");

  return quark;
}
