/* Esco-GDM: esco-user.c
 * 
 * Copyright (C) 2004 James M. Cape <jcape@ignore-your.tv>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <gtk/gtkicontheme.h>
#include <gtk/gtkstock.h>

#include "esco-user.h"
#include "esco-screen-manager.h"
#include "esco-gdm-config.h"
#include "ve-config.h"
#include "ve-misc.h"


/* ******** *
 *  Macros  *
 * ******** */

#define DEFAULT_ICON		"stock_person"
#define DEFAULT_ICON_SIZE	48
#define MAX_ICON_SIZE		192
#define MIN_ICON_SIZE		16


/* ************** *
 *  Enumerations  *
 * ************** */

enum
{
  PROP_0,
  NAME,
  DISPLAY_NAME,
  UID,
  HOME_DIR,
  SCREENS,
  ICON,
  ICON_SIZE
};


/* ************ *
 *  Structures  *
 * ************ */

struct _EscoUser
{
  GObject parent;

  gchar *name;
  gchar *display_name;
  gchar *home_dir;
  gulong uid;
  gchar *filename;

  GList *screens;

  GdkPixbuf *orig_icon;
  GdkPixbuf *icon;
  gint icon_size;

  guint screen_changed_id;
  guint screen_closed_id;
  guint theme_changed_id;

  gboolean stock_icon:1;
};

struct _EscoUserClass
{
  GObjectClass parent_class;
};


/* ******************* *
 *  GType Declaration  *
 * ******************* */

G_DEFINE_TYPE (EscoUser, esco_user, G_TYPE_OBJECT);


/* ******************* *
 *  Utility Functions  *
 * ******************* */

static inline gboolean
check_user_shell (const gchar * pwshell)
{
  static GHashTable *shells = NULL;

  if (shells == NULL)
    {
      gchar *usershell;

      shells = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

      setusershell ();

      usershell = getusershell ();

      while (usershell != NULL)
	{
	  g_hash_table_insert (shells, g_strdup (usershell), GUINT_TO_POINTER (TRUE));
	  usershell = getusershell ();
	}

      endusershell ();
    }

  return (g_hash_table_lookup (shells, pwshell) != NULL);
}


static void
update_overlay (EscoUser * user)
{
  g_object_unref (user->icon);

  if (user->screens != NULL)
    {
      GdkPixbuf *overlay, *img;
      gint dest_width, dest_height, img_width, img_height;

      user->icon = gdk_pixbuf_copy (user->orig_icon);

      dest_width = gdk_pixbuf_get_width (user->icon);
      dest_height = gdk_pixbuf_get_height (user->icon);

      img = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), "gnome-fs-desktop",
				      ((gdouble) MAX (dest_width, dest_height) / 2), 0, NULL);

      img_width = gdk_pixbuf_get_width (img);
      img_height = gdk_pixbuf_get_height (img);

      overlay = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE,
				gdk_pixbuf_get_bits_per_sample (user->icon),
				dest_width, dest_height);
      gdk_pixbuf_fill (overlay, 0x00000000);

      gdk_pixbuf_copy_area (img, 0, 0, img_width, img_height, overlay,
			    (dest_width - img_width), (dest_height - img_height));
      g_object_unref (img);

      gdk_pixbuf_composite (overlay, user->icon, 0, 0, dest_width, dest_height,
			    0.0, 0.0, 1.0, 1.0, GDK_INTERP_BILINEAR, 255);
      g_object_unref (overlay);
    }
  else
    {
      user->icon = g_object_ref (user->orig_icon);
    }

  g_object_notify (G_OBJECT (user), "icon");
}


static gboolean
check_user_file (const gchar * filename, gulong user, gssize max_file_size,
		 EscoRelaxPermsFlags relax)
{
  struct stat fileinfo;

  if (max_file_size > G_MAXINT)
    max_file_size = G_MAXINT;

  if (stat (filename, &fileinfo) < 0)
    return FALSE;

  if (G_UNLIKELY (!S_ISREG (fileinfo.st_mode)))
    return FALSE;

  if (G_UNLIKELY (fileinfo.st_uid != user))
    return FALSE;

  if (G_UNLIKELY ((fileinfo.st_mode & S_IWGRP) == S_IWGRP && (relax & ESCO_RELAX_PERMS_GROUP) == 0))
    return FALSE;

  if (G_UNLIKELY ((fileinfo.st_mode & S_IWOTH) == S_IWOTH && (relax & ESCO_RELAX_PERMS_OTHER) == 0))
    return FALSE;

  if (G_UNLIKELY (fileinfo.st_size > max_file_size))
    return FALSE;

  return TRUE;
}


static void
load_icon (EscoUser * user)
{
  gchar *filename;
  gssize max_user_size;
  EscoRelaxPermsFlags relax;

  /* Load the User's Pixbuf */
  max_user_size = esco_gdm_config_get_max_user_size ();
  relax = esco_gdm_config_get_relax_perms ();

  filename = g_build_filename (user->home_dir, ".face", NULL);
  if (check_user_file (filename, user->uid, max_user_size, relax))
    {
      user->orig_icon = gdk_pixbuf_new_from_file_at_size (filename, user->icon_size,
							  user->icon_size, NULL);
      user->filename = filename;
    }

  if (user->orig_icon == NULL)
    {
      filename = g_build_filename (user->home_dir, ".face.icon", NULL);
      if (check_user_file (filename, user->uid, max_user_size, relax))
	{
	  user->orig_icon = gdk_pixbuf_new_from_file_at_size (filename, user->icon_size,
							      user->icon_size, NULL);
	  user->filename = filename;
	}

      g_free (filename);

      if (user->orig_icon == NULL)
	{
	  VeConfig *cfg;

	  filename = g_build_filename (user->home_dir, ".gnome2", "gdm", NULL);
	  cfg = ve_config_new (filename);
	  g_free (filename);

	  filename = ve_config_get_string (cfg, "face/picture");
	  ve_config_destroy (cfg);
	  if (filename != NULL && check_user_file (filename, user->uid, max_user_size, relax))
	    {
	      user->orig_icon = gdk_pixbuf_new_from_file_at_size (filename, user->icon_size,
								  user->icon_size, NULL);
	    }

	  g_free (filename);

	  if (user->orig_icon == NULL)
	    {
	      filename = g_build_filename (user->home_dir, ".gnome", "photo", NULL);
	      if (check_user_file (filename, user->uid, max_user_size, relax))
		{
		  user->orig_icon = gdk_pixbuf_new_from_file_at_size (filename, user->icon_size,
								      user->icon_size, NULL);
		}

	      g_free (filename);

	      if (user->icon == NULL)
		{
		  gchar *face_dir;

		  face_dir = esco_gdm_config_get_global_face_dir ();
		  filename = g_build_filename (face_dir, user->name, NULL);
		  if (check_user_file (filename, user->uid, max_user_size, relax))
		    {
		      user->orig_icon = gdk_pixbuf_new_from_file_at_size (filename, user->icon_size,
									  user->icon_size, NULL);
		    }

		  g_free (filename);

		  if (user->orig_icon == NULL)
		    {
		      gchar *tmp;

		      tmp = g_strconcat (user->name, ".png", NULL);
		      filename = g_build_filename (face_dir, tmp, NULL);
		      g_free (tmp);

		      if (check_user_file (filename, user->uid, max_user_size, relax))
			{
			  user->orig_icon = gdk_pixbuf_new_from_file_at_size (filename,
									      user->icon_size,
									      user->icon_size,
									      NULL);
			}

		      if (user->orig_icon == NULL)
			{
			  user->orig_icon =
			    gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), DEFAULT_ICON,
						      user->icon_size, 0, NULL);

			  if (user->orig_icon == NULL)
			    {
			      user->orig_icon =
				gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
							  GTK_STOCK_MISSING_IMAGE,
							  user->icon_size, 0, NULL);
			    }

			  user->stock_icon = TRUE;
			}
		    }

		  g_free (face_dir);
		}
	    }
	}
    }
}


/* ***************** *
 *  Other Callbacks  *
 * ***************** */

static void
screen_closed_cb (EscoScreenManager * mgr, EscoScreen * screen, EscoUser * user)
{
  GList *screen_li;

  screen_li = g_list_find (user->screens, screen);
  if (screen_li != NULL)
    {
      user->screens = g_list_remove_link (user->screens, screen_li);
      g_object_notify (G_OBJECT (user), "screens");
      update_overlay (user);
    }
}


static void
screen_changed_cb (EscoScreenManager * mgr, EscoScreen * screen, EscoUser * user)
{
  GList *screen_li;
  gboolean dirty;

  screen_li = g_list_find (user->screens, screen);
  if (screen_li != NULL)
    {
      if (screen->username == NULL || strcmp (screen->username, user->name) != 0)
	user->screens = g_list_remove_link (user->screens, screen_li);

      dirty = TRUE;
    }
  else if (screen->username != NULL && strcmp (screen->username, user->name) == 0)
    {
      user->screens = g_list_prepend (user->screens, screen);
      dirty = TRUE;
    }
  else
    {
      dirty = FALSE;
    }

  if (dirty)
    {
      g_object_notify (G_OBJECT (user), "screens");
      update_overlay (user);
    }
}


static void
theme_changed_cb (GtkIconTheme * theme, EscoUser * user)
{
  if (user->stock_icon)
    {
      g_object_unref (user->orig_icon);
      user->orig_icon = gtk_icon_theme_load_icon (theme, DEFAULT_ICON, user->icon_size, 0, NULL);

      if (user->orig_icon == NULL)
	{
	  user->orig_icon = gtk_icon_theme_load_icon (theme, GTK_STOCK_MISSING_IMAGE,
						      user->icon_size, 0, NULL);
	}
    }

  update_overlay (user);
}


/* ******************* *
 *  GObject Functions  *
 * ******************* */

static GObject *
esco_user_constructor (GType type, guint n_params, GObjectConstructParam * params)
{
  GObject *object;
  EscoUser *user;

  object = (*G_OBJECT_CLASS (esco_user_parent_class)->constructor) (type, n_params, params);
  user = ESCO_USER (object);

  user->screens = esco_screen_manager_list_screens_for_user (esco_screen_manager_get_default (),
							     user->name);
  load_icon (user);
  update_overlay (user);

  return object;
}


static void
esco_user_set_property (GObject * object, guint id, const GValue * value, GParamSpec * pspec)
{
  EscoUser *user = ESCO_USER (object);

  switch (id)
    {
    case NAME:
      user->name = g_value_dup_string (value);
      break;
    case DISPLAY_NAME:
      user->display_name = g_value_dup_string (value);
      break;
    case HOME_DIR:
      user->home_dir = g_value_dup_string (value);
      break;
    case UID:
      user->uid = g_value_get_ulong (value);
      break;
    case ICON_SIZE:
      user->icon_size = g_value_get_int (value);

      if (user->orig_icon != NULL)
	{
	  g_object_unref (user->orig_icon);
	  load_icon (user);
	  update_overlay (user);
	  g_object_notify (object, "icon");
	}
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}


static void
esco_user_get_property (GObject * object, guint id, GValue * value, GParamSpec * pspec)
{
  EscoUser *user = ESCO_USER (object);

  switch (id)
    {
    case NAME:
      g_value_set_string (value, user->name);
      break;
    case DISPLAY_NAME:
      g_value_set_string (value, user->display_name);
      break;
    case HOME_DIR:
      g_value_set_string (value, user->home_dir);
      break;
    case SCREENS:
      {
	GList *list;
	GValueArray *ar;

	ar = g_value_array_new (g_list_length (user->screens));

	for (list = user->screens; list != NULL; list = list->next)
	  {
	    GValue tmp_value = { 0 };

	    g_value_init (&tmp_value, ESCO_TYPE_SCREEN);
	    g_value_set_boxed (&tmp_value, list->data);
	    g_value_array_prepend (ar, &tmp_value);
	    g_value_unset (&tmp_value);
	  }

	g_value_take_boxed (value, ar);
      }
      break;
    case UID:
      g_value_set_ulong (value, user->uid);
      break;
    case ICON:
      g_value_set_object (value, user->icon);
      break;
    case ICON_SIZE:
      g_value_set_int (value, user->icon_size);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}


static void
esco_user_dispose (GObject * object)
{
  EscoUser *user;
  EscoScreenManager *mgr;

  user = ESCO_USER (object);
  mgr = esco_screen_manager_get_default ();

  g_signal_handler_disconnect (mgr, user->screen_changed_id);
  g_signal_handler_disconnect (mgr, user->screen_closed_id);
  g_signal_handler_disconnect (gtk_icon_theme_get_default (), user->theme_changed_id);

  g_object_unref (user->orig_icon);
  g_object_unref (user->icon);

  if (G_OBJECT_CLASS (esco_user_parent_class)->dispose != NULL)
    (*G_OBJECT_CLASS (esco_user_parent_class)->dispose) (object);
}


static void
esco_user_finalize (GObject * object)
{
  EscoUser *user = ESCO_USER (object);

  g_free (user->name);
  g_free (user->display_name);
  g_free (user->home_dir);

  if (G_OBJECT_CLASS (esco_user_parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS (esco_user_parent_class)->finalize) (object);
}


/* ***************** *
 *  GType Functions  *
 * ***************** */

static void
esco_user_class_init (EscoUserClass * class)
{
  GObjectClass *object_class;

  object_class = G_OBJECT_CLASS (class);

  object_class->constructor = esco_user_constructor;
  object_class->set_property = esco_user_set_property;
  object_class->get_property = esco_user_get_property;
  object_class->dispose = esco_user_dispose;
  object_class->finalize = esco_user_finalize;

  g_object_class_install_property (object_class, NAME,
				   g_param_spec_string ("name", "Name",
							"The user this object represents.", NULL,
							(G_PARAM_READWRITE
							 | G_PARAM_CONSTRUCT_ONLY)));
  g_object_class_install_property (object_class, DISPLAY_NAME,
				   g_param_spec_string ("display-name", "Display Name",
							"The string to use in a user interface for "
							"this user.", NULL,
							(G_PARAM_READWRITE
							 | G_PARAM_CONSTRUCT_ONLY)));
  g_object_class_install_property (object_class, HOME_DIR,
				   g_param_spec_string ("home-directory", "Home Directory",
							"This user's home directory.", NULL,
							(G_PARAM_READWRITE
							 | G_PARAM_CONSTRUCT_ONLY)));
  g_object_class_install_property (object_class, UID,
				   g_param_spec_ulong ("user-id", "User ID",
						       "This user's numerical ID.",
						       0, G_MAXULONG, 0,
						       (G_PARAM_READWRITE
							| G_PARAM_CONSTRUCT_ONLY)));

  g_object_class_install_property (object_class, SCREENS,
				   g_param_spec_value_array ("screens", "Screens",
							     "A value array containing the user's "
							     "screens",
							     g_param_spec_boxed ("screen", "Screen",
										 "A single screen",
										 ESCO_TYPE_SCREEN,
										 G_PARAM_READABLE),
							     G_PARAM_READABLE));

  g_object_class_install_property (object_class, ICON,
				   g_param_spec_object ("icon", "Icon",
							"A pixbuf of this user's icon.",
							GDK_TYPE_PIXBUF, G_PARAM_READABLE));
  g_object_class_install_property (object_class, ICON_SIZE,
				   g_param_spec_int ("icon-size", "Icon Size",
						     "The size of this icon.", MIN_ICON_SIZE,
						     MAX_ICON_SIZE, DEFAULT_ICON_SIZE,
						     (G_PARAM_READWRITE | G_PARAM_CONSTRUCT)));
}


static void
esco_user_init (EscoUser * user)
{
  EscoScreenManager *mgr;

  mgr = esco_screen_manager_get_default ();

  user->screen_changed_id = g_signal_connect (mgr, "screen-changed",
					      G_CALLBACK (screen_changed_cb), user);
  user->screen_closed_id = g_signal_connect (mgr, "screen-closed", G_CALLBACK (screen_closed_cb),
					     user);
  user->theme_changed_id = g_signal_connect (gtk_icon_theme_get_default (), "changed",
					     G_CALLBACK (theme_changed_cb), user);
}


/* ************************************************************************** *
 *  Public API                                                                *
 * ************************************************************************** */

EscoUser *
esco_user_new (const gchar * user_name, gint icon_size)
{
  struct passwd *pwent;

  g_return_val_if_fail (user_name != NULL && user_name[0] != '\0', NULL);

  pwent = getpwnam (user_name);

  if (pwent == NULL)
    {
      g_warning ("Could not load user data for `%s'.", user_name);
      return NULL;
    }

  return esco_user_new_from_passwd (pwent, icon_size);
}


EscoUser *
esco_user_new_from_passwd (const struct passwd * pwent, gint icon_size)
{
  EscoUser *retval;
  gchar *display_name;

  g_return_val_if_fail (pwent != NULL, NULL);

  if (pwent->pw_gecos != NULL)
    {
      display_name = g_strdup (pwent->pw_gecos);

      g_strstrip (display_name);

      if (display_name[0] == '\0')
	{
	  g_free (display_name);
	  display_name = g_strdup (pwent->pw_name);
	}
      else
	{
	  gchar *ptr;

	  ptr = strchr (display_name, ',');

	  if (ptr != NULL)
	    *ptr = '\0';
	}
    }
  else
    {
      display_name = g_strdup (pwent->pw_name);
    }

  retval = g_object_new (ESCO_TYPE_USER, "name", pwent->pw_name, "display-name", display_name,
			 "user-id", (gulong) pwent->pw_uid, "icon-size", icon_size,
			 "home-directory", pwent->pw_dir, NULL);

  g_free (display_name);

  return retval;
}


void
esco_user_set_icon_size (EscoUser *user, gint icon_size)
{
  gint real_icon_size;

  g_return_if_fail (ESCO_IS_USER (user));

  real_icon_size = CLAMP (icon_size, MIN_ICON_SIZE, MAX_ICON_SIZE);

  if (real_icon_size == user->icon_size)
    return;

  user->icon_size = real_icon_size;
  load_icon (user);
  update_overlay (user);
}


GdkPixbuf *
esco_user_get_icon (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), NULL);

  return user->icon;
}


G_CONST_RETURN gchar *
esco_user_get_name (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), NULL);

  return user->name;
}


G_CONST_RETURN gchar *
esco_user_get_display_name (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), NULL);

  return user->display_name;
}


G_CONST_RETURN gchar *
esco_user_get_home_directory (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), NULL);

  return user->home_dir;
}


G_CONST_RETURN GList *
esco_user_get_screens (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), NULL);

  return user->screens;
}


gulong
esco_user_get_uid (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), 0);

  return user->uid;
}


gint
esco_user_get_icon_size (EscoUser *user)
{
  g_return_val_if_fail (ESCO_IS_USER (user), 0);

  return user->icon_size;
}


gboolean
esco_user_is_visible (const struct passwd *pwent)
{
  GList *screens;

  g_return_val_if_fail (pwent != NULL, FALSE);

  screens = esco_screen_manager_list_screens_for_user (esco_screen_manager_get_default (),
						       pwent->pw_name);
  if (screens != NULL)
    {
      g_list_foreach (screens, (GFunc) esco_screen_unref, NULL);
      g_list_free (screens);

      return TRUE;
    }

  return (strcmp (pwent->pw_name, "nobody") != 0
          && check_user_shell (pwent->pw_shell)
	  && (pwent->pw_uid != 0
	      || esco_gdm_config_get_allow_root ())
	  && pwent->pw_uid >= esco_gdm_config_get_minimal_uid ()
	  && !esco_gdm_config_is_user_excluded (pwent->pw_name));
}


gboolean
esco_user_name_is_visible (const gchar *username)
{
  struct passwd *pwent;

  g_return_val_if_fail (!ve_string_empty (username), FALSE);

  pwent = getpwnam (username);

  if (pwent == NULL)
    return FALSE;

  return esco_user_is_visible (pwent);
}
