/* Escobar Libraries: escofilechooserentry.c
 * 
 * Copyright (c) 2004 James M. Cape <jcape@ignore-your.tv>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <string.h>

#include <glib/gi18n.h>

#include <gtk/gtkcelllayout.h>
#include <gtk/gtkcellrendererpixbuf.h>
#include <gtk/gtkcomboboxentry.h>
#include <gtk/gtkfilechooserdialog.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkicontheme.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtkstock.h>
#include <gtk/gtktogglebutton.h>

#include "gtkfilechooserprivate.h"
#include "gtkfilechooserutils.h"
#include "escofilechooserentry.h"


/* **************** *
 *  Private Macros  *
 * **************** */

#define ESCO_FILE_CHOOSER_ENTRY_GET_PRIVATE(object) (ESCO_FILE_CHOOSER_ENTRY ((object))->_priv)

#define DEFAULT_TITLE		N_("Select a File")
#define DEFAULT_MAX_HISTORY_SIZE	10

/* ********************** *
 *  Private Enumerations  *
 * ********************** */

/* Property IDs */
enum
{
  PROP_0,

  PROP_TITLE,
  PROP_ACTIVE,
  PROP_HISTORY,
  PROP_MAX_HISTORY_SIZE
};

/* GtkTreeModel Columns */
enum
{
  TEXT_COLUMN,
  HASH_COLUMN,
  NUM_COLUMNS
};


/* ******************** *
 *  Private Structures  *
 * ******************** */

struct _EscoFileChooserEntryPrivate
{
  GtkWidget *dialog;
  GtkWidget *combo;
  GtkWidget *button;
  GtkWidget *image;

  gchar *filesystem;
  guint max_size;
};


/* ******************* *
 *  GType Declaration  *
 * ******************* */

G_DEFINE_TYPE_WITH_CODE (EscoFileChooserEntry, esco_file_chooser_entry, GTK_TYPE_HBOX, { \
  G_IMPLEMENT_INTERFACE (GTK_TYPE_FILE_CHOOSER, _gtk_file_chooser_delegate_iface_init) \
});


/* ********************** *
 *  Sub-Widget Callbacks  *
 * ********************** */

static void
dialog_response_cb (GtkFileChooser * dialog, gint response, EscoFileChooserEntry * entry)
{
  gboolean hide_dialog;

  if (response == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename;

      filename = gtk_file_chooser_get_filename (dialog);

      if (filename != NULL)
	{
	  hide_dialog = TRUE;
	  esco_file_chooser_entry_add_history (entry, filename);
	  g_free (filename);
	}
      else
	{
	  hide_dialog = FALSE;
	}
    }
  else
    {
      hide_dialog = TRUE;
    }

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (entry->_priv->button), !hide_dialog);
}


static void
button_toggled_cb (GtkToggleButton * button, EscoFileChooserEntry * entry)
{
  if (gtk_toggle_button_get_active (button))
    {
      if (!GTK_WIDGET_VISIBLE (entry->_priv->dialog))
	{
	  GtkWidget *toplevel;

	  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (entry));

	  if (GTK_WIDGET_TOPLEVEL (toplevel) && GTK_IS_WINDOW (toplevel))
	    {
	      if (GTK_WINDOW (toplevel) !=
		  gtk_window_get_transient_for (GTK_WINDOW (entry->_priv->dialog)))
		{
		  gtk_window_set_transient_for (GTK_WINDOW (entry->_priv->dialog),
						GTK_WINDOW (toplevel));
		}
	      gtk_window_set_modal (GTK_WINDOW (entry->_priv->dialog),
				    gtk_window_get_modal (GTK_WINDOW (toplevel)));
	    }
	}

      gtk_window_present (GTK_WINDOW (entry->_priv->dialog));
    }
  else
    {
      gtk_widget_hide (entry->_priv->dialog);
    }
}


static void
combo_size_request_cb (GtkWidget * combo, GtkRequisition * req, GtkWidget * button)
{
  gtk_widget_set_size_request (button, -1, req->height);
}


static void
render_mime_icon (GtkCellLayout *layout, GtkCellRenderer *cell, GtkTreeModel *model,
		  GtkTreeIter *iter, EscoFileChooserEntry *entry)
{
  GdkPixbuf *pixbuf;
  gchar *filename;

  filename = NULL;
  gtk_tree_model_get (model, iter, TEXT_COLUMN, &filename, -1);

  if (filename != NULL)
    {
      GtkFileSystem *fs;
      GtkFilePath *path;

      fs = _gtk_file_chooser_get_file_system (GTK_FILE_CHOOSER (entry->_priv->dialog));

      path = gtk_file_system_filename_to_path (fs, filename);
      switch (gtk_file_chooser_get_action (GTK_FILE_CHOOSER (entry->_priv->dialog)))
	{
	case GTK_FILE_CHOOSER_ACTION_OPEN:
	case GTK_FILE_CHOOSER_ACTION_SAVE:
	  pixbuf = gtk_file_system_render_icon (fs, path, GTK_WIDGET (entry), 16, NULL);
	  break;
	case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
	case GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER:
	  {
	    GtkFileSystemVolume *volume;

	    volume = gtk_file_system_get_volume_for_path (fs, path);
	    pixbuf = gtk_file_system_volume_render_icon (fs, volume, GTK_WIDGET (entry), 16, NULL);
	    gtk_file_system_volume_free (fs, volume);
	  }
	  break;
	default:
	  g_assert_not_reached ();
	  pixbuf = NULL;
	  break;
	}
      gtk_file_path_free (path);
    }
  else
    {
      pixbuf = NULL;
    }

  g_object_set (cell, "pixbuf", pixbuf, NULL);

  if (pixbuf != NULL)
    g_object_unref (pixbuf);
}


/* ********************* *
 *  GtkObject Functions  *
 * ********************* */

static void
esco_file_chooser_entry_destroy (GtkObject * object)
{
  EscoFileChooserEntryPrivate *priv;

  priv = ESCO_FILE_CHOOSER_ENTRY_GET_PRIVATE (object);

  if (priv->dialog != NULL)
    gtk_widget_destroy (priv->dialog);

  if (GTK_OBJECT_CLASS (esco_file_chooser_entry_parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS (esco_file_chooser_entry_parent_class)->destroy) (object);
}


/* ******************* *
 *  GObject Functions  *
 * ******************* */

static GObject *
esco_file_chooser_entry_constructor (GType type, guint n_params, GObjectConstructParam * params)
{
  GObject *object;
  EscoFileChooserEntryPrivate *priv;
  GtkListStore *store;
  GtkCellRenderer *cell;

  object = (*G_OBJECT_CLASS (esco_file_chooser_entry_parent_class)->constructor) (type, n_params,
										  params);

  priv = ESCO_FILE_CHOOSER_ENTRY_GET_PRIVATE (object);

  priv->dialog = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
			       "file-system-backend", priv->filesystem, NULL);
  gtk_dialog_add_button (GTK_DIALOG (priv->dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT);
  gtk_dialog_add_button (GTK_DIALOG (priv->dialog), GTK_STOCK_OK, GTK_RESPONSE_ACCEPT);
  g_signal_connect (priv->dialog, "response", G_CALLBACK (dialog_response_cb), object);
  g_object_add_weak_pointer (G_OBJECT (priv->dialog), (gpointer *) (&priv->dialog));

  g_free (priv->filesystem);
  priv->filesystem = NULL;

  store = gtk_list_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_UINT);
  priv->combo = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL (store), TEXT_COLUMN);
  g_object_unref (store);
  gtk_container_add (GTK_CONTAINER (object), priv->combo);
  gtk_widget_show (priv->combo);

  cell = gtk_cell_renderer_pixbuf_new ();
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (priv->combo), cell, FALSE);
  gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (priv->combo), cell,
				      (GtkCellLayoutDataFunc) render_mime_icon, object, NULL);
  gtk_cell_layout_reorder (GTK_CELL_LAYOUT (priv->combo), cell, 0);

  priv->button = gtk_toggle_button_new ();
  g_signal_connect (priv->combo, "size-request", G_CALLBACK (combo_size_request_cb), priv->button);
  g_signal_connect (priv->button, "clicked", G_CALLBACK (button_toggled_cb), object);
  gtk_box_pack_start (GTK_BOX (object), priv->button, FALSE, FALSE, 0);
  gtk_widget_show (priv->button);

  priv->image = gtk_image_new_from_stock (GTK_STOCK_OPEN, GTK_ICON_SIZE_MENU);
  gtk_container_add (GTK_CONTAINER (priv->button), priv->image);
  gtk_widget_show (priv->image);

  _gtk_file_chooser_set_delegate (GTK_FILE_CHOOSER (object), GTK_FILE_CHOOSER (priv->dialog));

  return object;
}


static void
esco_file_chooser_entry_set_property (GObject * object, guint id, const GValue * value,
				      GParamSpec * pspec)
{
  EscoFileChooserEntry *entry = ESCO_FILE_CHOOSER_ENTRY (object);

  switch (id)
    {
    case PROP_ACTIVE:
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (entry->_priv->button),
				    g_value_get_boolean (value));
      break;

    case PROP_MAX_HISTORY_SIZE:
      esco_file_chooser_entry_set_max_history_size (entry, g_value_get_uint (value));
      break;

    case PROP_HISTORY:
      {
	GValueArray *ar;
	GtkListStore *store;
	guint i;

	ar = g_value_get_boxed (value);
	store = GTK_LIST_STORE (gtk_combo_box_get_model (GTK_COMBO_BOX (entry->_priv->combo)));

	for (i = ar->n_values; i >= 0; i--)
	  {
	    const gchar *str;

	    str = g_value_get_string (g_value_array_get_nth (ar, i));

	    if (str != NULL)
	      {
		GtkTreeIter iter;

		gtk_list_store_prepend (store, &iter);
		gtk_list_store_set (store, &iter, TEXT_COLUMN, str, HASH_COLUMN, g_str_hash (str),
				    -1);
	      }
	  }
      }
      break;

    case GTK_FILE_CHOOSER_PROP_ACTION:
      switch (g_value_get_enum (value))
	{
	case GTK_FILE_CHOOSER_ACTION_OPEN:
	case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
	  gtk_image_set_from_stock (GTK_IMAGE (entry->_priv->image),
				    GTK_STOCK_OPEN, GTK_ICON_SIZE_MENU);
	  break;
	case GTK_FILE_CHOOSER_ACTION_SAVE:
	  gtk_image_set_from_stock (GTK_IMAGE (entry->_priv->image),
				    GTK_STOCK_NEW, GTK_ICON_SIZE_MENU);
	  break;
	case GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER:
	  gtk_image_set_from_stock (GTK_IMAGE (entry->_priv->image),
				    GTK_STOCK_NEW, GTK_ICON_SIZE_MENU);
	  break;
	default:
	  g_assert_not_reached ();
	  break;
	}
    case PROP_TITLE:
    case GTK_FILE_CHOOSER_PROP_FILTER:
    case GTK_FILE_CHOOSER_PROP_LOCAL_ONLY:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET_ACTIVE:
    case GTK_FILE_CHOOSER_PROP_USE_PREVIEW_LABEL:
    case GTK_FILE_CHOOSER_PROP_EXTRA_WIDGET:
    case GTK_FILE_CHOOSER_PROP_SHOW_HIDDEN:
      g_object_set_property (G_OBJECT (entry->_priv->dialog), pspec->name, value);
      break;

    case GTK_FILE_CHOOSER_PROP_FILE_SYSTEM_BACKEND:
      /* Construct-only */
      entry->_priv->filesystem = g_value_dup_string (value);
      break;

    case GTK_FILE_CHOOSER_PROP_SELECT_MULTIPLE:
      g_warning ("%s: Choosers of type `%s` do not support selecting multiple files.", G_STRFUNC,
		 G_OBJECT_TYPE_NAME (object));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}


static void
esco_file_chooser_entry_get_property (GObject * object, guint id, GValue * value,
				      GParamSpec * pspec)
{
  EscoFileChooserEntryPrivate *priv = ESCO_FILE_CHOOSER_ENTRY_GET_PRIVATE (object);

  switch (id)
    {
    case PROP_ACTIVE:
      g_value_set_boolean (value, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->button)));
      break;

    case PROP_MAX_HISTORY_SIZE:
      g_value_set_uint (value, priv->max_size);
      break;

    case PROP_HISTORY:
      {
	GtkTreeModel *model;
	GtkTreeIter iter;

	model = gtk_combo_box_get_model (GTK_COMBO_BOX (priv->combo));

	if (gtk_tree_model_get_iter_first (model, &iter))
	  {
	    GValueArray *ar;
	    GValue tmp_value = { 0 };

	    ar = g_value_array_new (0);
	    g_value_init (&tmp_value, G_TYPE_STRING);

	    do
	      {
		gtk_tree_model_get_value (model, &iter, TEXT_COLUMN, &tmp_value);
		g_value_array_append (ar, &tmp_value);
		g_value_reset (&tmp_value);
	      }
	    while (gtk_tree_model_iter_next (model, &iter));

	    g_value_take_boxed (value, ar);
	  }
	else
	  {
	    g_value_set_boxed (value, NULL);
	  }
      }
      break;

    case PROP_TITLE:
      /* GtkFileChooser Properties */
    case GTK_FILE_CHOOSER_PROP_ACTION:
    case GTK_FILE_CHOOSER_PROP_FILE_SYSTEM_BACKEND:
    case GTK_FILE_CHOOSER_PROP_FILTER:
    case GTK_FILE_CHOOSER_PROP_LOCAL_ONLY:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET_ACTIVE:
    case GTK_FILE_CHOOSER_PROP_USE_PREVIEW_LABEL:
    case GTK_FILE_CHOOSER_PROP_EXTRA_WIDGET:
    case GTK_FILE_CHOOSER_PROP_SELECT_MULTIPLE:
    case GTK_FILE_CHOOSER_PROP_SHOW_HIDDEN:
      g_object_get_property (G_OBJECT (priv->dialog), pspec->name, value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}


/* ***************** *
 *  GType Functions  *
 * ***************** */

static void
esco_file_chooser_entry_class_init (EscoFileChooserEntryClass * class)
{
  GObjectClass *gobject_class;
  GtkObjectClass *gtkobject_class;
  GtkWidgetClass *widget_class;

  gobject_class = G_OBJECT_CLASS (class);
  gtkobject_class = GTK_OBJECT_CLASS (class);
  widget_class = GTK_WIDGET_CLASS (class);

  gobject_class->constructor = esco_file_chooser_entry_constructor;
  gobject_class->set_property = esco_file_chooser_entry_set_property;
  gobject_class->get_property = esco_file_chooser_entry_get_property;

  gtkobject_class->destroy = esco_file_chooser_entry_destroy;

  g_object_class_install_property (gobject_class, PROP_TITLE,
				   g_param_spec_string ("title", "Title",
							"The title of the file chooser dialog.",
							_(DEFAULT_TITLE), G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, PROP_ACTIVE,
				   g_param_spec_boolean ("active", "Active",
							 "Whether the browse dialog is visible or "
							 "not.", FALSE, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, PROP_MAX_HISTORY_SIZE,
				   g_param_spec_uint ("max-history-size", "Maximum History Size",
						      "The maximum size of the history to store.",
						      0, G_MAXINT, DEFAULT_MAX_HISTORY_SIZE,
						      G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, PROP_HISTORY,
				   g_param_spec_value_array ("history", "History",
							     "The history of selected filenames.",
							     g_param_spec_string ("history-item",
										  "History Item",
										  "A previously "
										  "selected "
										  "filename", NULL,
										  G_PARAM_READWRITE),
							     G_PARAM_READWRITE));

  _gtk_file_chooser_install_properties (gobject_class);

  g_type_class_add_private (class, sizeof (EscoFileChooserEntryPrivate));
}


static void
esco_file_chooser_entry_init (EscoFileChooserEntry * entry)
{
  entry->_priv = G_TYPE_INSTANCE_GET_PRIVATE (entry, ESCO_TYPE_FILE_CHOOSER_ENTRY,
					      EscoFileChooserEntryPrivate);
}


/* ************************************************************************** *
 *  Public API                                                                *
 * ************************************************************************** */

/**
 * esco_file_chooser_entry_new:
 * @title: the title of the browse dialog.
 * 
 * Creates a new file-selecting entry widget.
 * 
 * Returns: a new entry widget.
 * 
 * Since: 2.6
 **/
GtkWidget *
esco_file_chooser_entry_new (const gchar * title)
{
  return g_object_new (ESCO_TYPE_FILE_CHOOSER_ENTRY, "title", title, NULL);
}


/**
 * esco_file_chooser_entry_set_title:
 * @entry: the entry widget to modify.
 * @title: the new browse dialog title.
 * 
 * Modifies the @title of the browse dialog used by @entry.
 * 
 * Since: 2.6
 **/
void
esco_file_chooser_entry_set_title (EscoFileChooserEntry * entry, const gchar * title)
{
  g_return_if_fail (ESCO_IS_FILE_CHOOSER_ENTRY (entry));

  gtk_window_set_title (GTK_WINDOW (entry->_priv->dialog), title);
  g_object_notify (G_OBJECT (entry), "title");
}


/**
 * esco_file_chooser_entry_get_title:
 * @entry: the entry widget to examine.
 * 
 * Retrieves the title of the browse dialog used by @entry. The returned value
 * should not be modified or freed.
 * 
 * Returns: a pointer to the browse dialog's title.
 * 
 * Since: 2.6
 **/
G_CONST_RETURN gchar *
esco_file_chooser_entry_get_title (EscoFileChooserEntry * entry)
{
  g_return_val_if_fail (ESCO_IS_FILE_CHOOSER_ENTRY (entry), NULL);

  return gtk_window_get_title (GTK_WINDOW (entry->_priv->dialog));
}


/**
 * esco_file_chooser_entry_set_max_max_size:
 * @entry: the entry widget to modify.
 * @max_size: the new maximum history size.
 * 
 * Sets the maximum length of the history pulldown of @entry to @max_size. If
 * the current history list size is greater than @max_size, than the list will
 * be truncated from the oldest entries to the newest. If @max_size is %0, then
 * @entry will not limit it's history list size.
 * 
 * Since: 2.6
 **/
void
esco_file_chooser_entry_set_max_history_size (EscoFileChooserEntry * entry, guint max_size)
{
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (ESCO_IS_FILE_CHOOSER_ENTRY (entry));

  entry->_priv->max_size = max_size;
  g_object_notify (G_OBJECT (entry), "max-history-size");

  model = gtk_combo_box_get_model (GTK_COMBO_BOX (entry->_priv->combo));

  if (gtk_tree_model_get_iter_first (model, &iter))
    {
      guint n_items = 0;

      do
	{
	  n_items++;

	  if (n_items > max_size)
	    {
	      gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
	    }
	}
      while (gtk_tree_model_iter_next (model, &iter));
    }
}


/**
 * esco_file_chooser_entry_add_history:
 * @entry: the entry widget to examine.
 * @filename: a full path to a file to the history pulldown.
 * 
 * Adds a path to the pulldown history list for @entry. The @filename must be a
 * valid path to an existing file or directory. When the @entry already has a
 * maximum number of entries (set by the "max-history-size" property or the
 * esco_file_chooser_entry_set_max_size() function), and the history has reached
 * that limit, the oldest entry will be removed. This item will always be 
 * 
 * Since: 2.6
 **/
void
esco_file_chooser_entry_add_history (EscoFileChooserEntry * entry, const gchar * filename)
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  guint hash, list_size;
  gboolean duplicate;

  g_return_if_fail (ESCO_IS_FILE_CHOOSER_ENTRY (entry));
  g_return_if_fail (filename != NULL && filename[0] != '\0');

  switch (gtk_file_chooser_get_action (GTK_FILE_CHOOSER (entry->_priv->dialog)))
    {
    case GTK_FILE_CHOOSER_ACTION_OPEN:
      g_return_if_fail (g_file_test (filename, G_FILE_TEST_EXISTS));
      break;
    case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
      g_return_if_fail (g_file_test (filename, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)));
      break;
    default:
      break;
    }

  duplicate = FALSE;
  list_size = 0;
  hash = g_str_hash (filename);

  model = gtk_combo_box_get_model (GTK_COMBO_BOX (entry->_priv->combo));
  if (gtk_tree_model_get_iter_first (model, &iter))
    {
      do
	{
	  guint tmp_hash = 0;

	  gtk_tree_model_get (model, &iter, HASH_COLUMN, tmp_hash, -1);

	  if (hash == tmp_hash)
	    {
	      gchar *tmp_filename = NULL;

	      gtk_tree_model_get (model, &iter, TEXT_COLUMN, &tmp_filename, -1);

	      g_assert (tmp_filename != NULL);

	      if (strcmp (filename, tmp_filename) == 0)
		duplicate = TRUE;

	      g_free (tmp_filename);
	    }

	  list_size++;
	}
      while (!duplicate && gtk_tree_model_iter_next (model, &iter));
    }

  /* If @filename is a duplicate, or the list is full, delete the last-visited iter */
  if (duplicate || (entry->_priv->max_size > 0 && list_size == entry->_priv->max_size))
    gtk_list_store_remove (GTK_LIST_STORE (model), &iter);

  gtk_list_store_prepend (GTK_LIST_STORE (model), &iter);
  gtk_list_store_set (GTK_LIST_STORE (model), &iter, TEXT_COLUMN, filename, HASH_COLUMN, hash, -1);
  gtk_combo_box_set_active (GTK_COMBO_BOX (entry->_priv->combo), 0);

  g_object_notify (G_OBJECT (entry), "history");
}


/**
 * esco_file_chooser_entry_set_history:
 * @entry: the entry widget to modify.
 * @history: a list of local path strings.
 * 
 * Overwrites any existing history lists in @entry with elements of @history.
 * The given @history should be sorted from most-recent to least-recent. Also,
 * if a max-history-size is set on @entry, only the that many items will be
 * copied from the start of @history.
 * 
 * Since: 2.6
 **/
void
esco_file_chooser_entry_set_history (EscoFileChooserEntry * entry, GSList * history)
{
  GtkListStore *store;
  GtkFileChooserAction action;
  guint n_items;

  g_return_if_fail (ESCO_IS_FILE_CHOOSER_ENTRY (entry));

  action = gtk_file_chooser_get_action (GTK_FILE_CHOOSER (entry->_priv->dialog));
  store = GTK_LIST_STORE (gtk_combo_box_get_model (GTK_COMBO_BOX (entry->_priv->combo)));
  gtk_list_store_clear (store);

  n_items = 0;
  while (history != NULL && (entry->_priv->max_size > 0 || n_items < entry->_priv->max_size))
    {
      switch (action)
	{
	case GTK_FILE_CHOOSER_ACTION_OPEN:
	  if (!g_file_test (history->data, G_FILE_TEST_EXISTS))
	    continue;
	  break;
	case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
	  if (!g_file_test (history->data, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)))
	    continue;
	  break;
	default:
	  break;
	}

      if (history->data != NULL)
	{
	  GtkTreeIter iter;

	  gtk_list_store_append (store, &iter);
	  gtk_list_store_set (store, &iter, TEXT_COLUMN, history->data,
			      HASH_COLUMN, g_str_hash (history->data), -1);
	  n_items++;
	}

      history = history->next;
    }

  g_object_notify (G_OBJECT (entry), "history");
}


/**
 * esco_file_chooser_entry_get_history:
 * @entry: the entry widget to examine.
 * 
 * Overwrites any existing history lists in @entry with the contents of @history.
 * 
 * Since: 2.6
 **/
GSList *
esco_file_chooser_entry_get_history (EscoFileChooserEntry * entry)
{
  GSList *retval;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_val_if_fail (ESCO_IS_FILE_CHOOSER_ENTRY (entry), NULL);

  retval = NULL;
  model = gtk_combo_box_get_model (GTK_COMBO_BOX (entry->_priv->combo));

  if (gtk_tree_model_get_iter_first (model, &iter))
    {
      do
	{
	  gchar *filename;

	  filename = NULL;
	  gtk_tree_model_get (model, &iter, TEXT_COLUMN, &filename, -1);

	  if (filename != NULL)
	    retval = g_slist_prepend (retval, filename);
	}
      while (gtk_tree_model_iter_next (model, &iter));
    }

  return g_slist_reverse (retval);
}
