/* Escobar Libraries: escoimagebutton.c
 * 
 * Copyright (c) 2004 James M. Cape <jcape@ignore-your.tv>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <string.h>

#include <glib/gi18n.h>

#include <gtk/gtkicontheme.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkfilechooserdialog.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkstock.h>
#include <gtk/gtkvbox.h>

#include "gtkfilechooserutils.h"
#include "gtkfilechooserprivate.h"
#include "escoimagebutton.h"


/* **************** *
 *  Private Macros  *
 * **************** */

#define ESCO_IMAGE_BUTTON_GET_PRIVATE(object)	(ESCO_IMAGE_BUTTON ((object))->_priv)
#define IMAGE_IS_VALID_KEY			(image_is_valid_key_quark ())

#define DEFAULT_TITLE				N_("Select a Picture")
#define DEFAULT_LABEL				N_("No Image")
#define DEFAULT_MIN_SIZE			48
#define BUFFER_SIZE				8192


/* ********************** *
 *  Private Enumerations  *
 * ********************** */

/* Property IDs */
enum
{
  PROP_0,

  PROP_TITLE,
  PROP_MIN_SIZE,
  PROP_SHOW_LABEL
};


/* ******************** *
 *  Private Structures  *
 * ******************** */

struct _EscoImageButtonPrivate
{
  GtkWidget *dialog;

  GtkWidget *preview_image;
  GtkWidget *preview_label;

  GtkWidget *image;
  GtkWidget *label;

  guint min_size;
  gchar *filesystem;

  gint src_width;
  gint src_height;

  gint image_width;
  gint image_height;

  guint update_id;
};

typedef struct
{
  gint dest_width;
  gint dest_height;
  gint *src_width;
  gint *src_height;
  gchar **type;
}
PixbufLoaderData;


/* ******************* *
 *  Utility Functions  *
 * ******************* */

static void
loader_size_prepared_cb (GdkPixbufLoader * loader, gint width, gint height, PixbufLoaderData * data)
{
  GdkPixbufFormat *format;
  gchar *name;
  gboolean is_svg;

  format = gdk_pixbuf_loader_get_format (loader);

  /*
   * FIXME: This can bit can be optimized out if/when
   * 
   * is fixed.
   */
  name = gdk_pixbuf_format_get_name (format);
  is_svg = (strcmp (name, "svg") == 0);
  g_free (name);

  if (data->type != NULL)
    {
      gchar **mime_types;
      guint i;

      mime_types = gdk_pixbuf_format_get_mime_types (format);
      g_assert (mime_types != NULL);
      *(data->type) = mime_types[0];

      if (mime_types[0] != NULL)
	{
	  for (i = 1; mime_types[i] != NULL; i++)
	    g_free (mime_types[i]);
	}

      g_free (mime_types);
    }

  if (is_svg)
    {
      if (data->src_width != NULL)
	*(data->src_width) = -1;
      if (data->src_height != NULL)
	*(data->src_height) = -1;

      if ((gdouble) height * (gdouble) data->dest_width >
	  (gdouble) width * (gdouble) data->dest_height)
	{
	  width = 0.5 + (gdouble) width *(gdouble) data->dest_height / (gdouble) height;
	  height = data->dest_height;
	}
      else
	{
	  height = 0.5 + (gdouble) height *(gdouble) data->dest_width / (gdouble) width;
	  width = data->dest_width;
	}
    }
  else
    {
      if (data->src_width != NULL)
	*(data->src_width) = width;
      if (data->src_height != NULL)
	*(data->src_height) = height;

      if (width > data->dest_width || height > data->dest_height)
	{
	  if ((gdouble) height * (gdouble) data->dest_width >
	      (gdouble) width * (gdouble) data->dest_height)
	    {
	      width = 0.5 + (gdouble) width *(gdouble) data->dest_height / (gdouble) height;
	      height = data->dest_height;
	    }
	  else
	    {
	      height = 0.5 + (gdouble) height *(gdouble) data->dest_width / (gdouble) width;
	      width = data->dest_width;
	    }
	}
    }

  gdk_pixbuf_loader_set_size (loader, width, height);
}


static GdkPixbuf *
util_load_pixbuf_for_size (const gchar * filename, gint dest_width, gint dest_height,
			   gint * src_width, gint * src_height, gchar **type, GError **error)
{
  GdkPixbuf *pixbuf;
  GdkPixbufLoader *loader;
  gint fd;
  guchar buffer[BUFFER_SIZE];
  ssize_t result;
  PixbufLoaderData data;

  fd = open (filename, O_RDONLY);

  if (fd < 0)
    return NULL;

  if (src_width != NULL)
    *src_width = -1;
  if (src_height != NULL)
    *src_height = -1;
  if (type != NULL)
    *type = NULL;

  data.dest_width = dest_width;
  data.dest_height = dest_height;
  data.src_width = src_width;
  data.src_height = src_height;
  data.type = type;

  loader = gdk_pixbuf_loader_new ();
  g_signal_connect (loader, "size-prepared", G_CALLBACK (loader_size_prepared_cb), &data);

  do
    {
      result = read (fd, buffer, sizeof (buffer));

      if (result > 0)
	{
	  if (!gdk_pixbuf_loader_write (loader, buffer, result, error))
	    {
	      result = -1;
	      break;
	    }
	}
    }
  while (result > 0);

  close (fd);
  gdk_pixbuf_loader_close (loader, NULL);

  if (result == 0)
    {
      pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);

      if (pixbuf != NULL)
	g_object_ref (pixbuf);
    }
  else
    {
      pixbuf = NULL;

      if (error != NULL && *error == NULL)
	g_set_error (error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
		     _("Failed to load image '%s': reason not known, probably a corrupt image file"),
		     filename);
    }

  if (pixbuf == NULL)
    {
      if (src_width != NULL)
	*src_width = -1;
      if (src_height != NULL)
	*src_height = -1;

      /* Must free because of alloc from the "size-prepared" cb */
      if (type != NULL)
	{
	  g_free (*type);
	  *type = NULL;
	}
    }

  g_object_unref (loader);

  return pixbuf;
}


static gboolean
try_set_button_image (EscoImageButton *button, const gchar *filename, GError **error)
{
  GdkPixbuf *pixbuf;
  gint width, height;

  width = MAX ((gint) button->_priv->min_size, button->_priv->image->allocation.width);
  height = MAX ((gint) button->_priv->min_size, button->_priv->image->allocation.height);

  if (filename != NULL)
    {
      if (width == 0 && height == 0)
	{
	  pixbuf = gdk_pixbuf_new_from_file (filename, error);
	  button->_priv->src_width = gdk_pixbuf_get_width (pixbuf);
	  button->_priv->src_height = gdk_pixbuf_get_height (pixbuf);
	}
      else
	{
	  pixbuf = util_load_pixbuf_for_size (filename, width, height,
					      &(button->_priv->src_width),
					      &(button->_priv->src_height), NULL, error);
	}
    }
  else
    {
      pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), "stock_placeholder-graphic",
					 16, 0, NULL);
    }

  gtk_image_set_from_pixbuf (GTK_IMAGE (button->_priv->image), pixbuf);

  if (pixbuf != NULL)
    g_object_unref (pixbuf);

  return (pixbuf != NULL);
}


static gboolean
reset_button_image (EscoImageButton *button)
{
  gchar *filename;

  filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (button->_priv->dialog));
  try_set_button_image (button, filename, NULL);
  g_free (filename);

  button->_priv->update_id = 0;
  return FALSE;
}


static void
reset_button_label (EscoImageButton *button)
{
  gchar *markup;
  gchar *filename;

  filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (button->_priv->dialog));

  if (filename != NULL)
    {
      gchar *basename, *basename_utf8;

      basename = g_path_get_basename (filename);
      g_free (filename);
      basename_utf8 = g_filename_to_utf8 (basename, -1, NULL, NULL, NULL);
      g_free (basename);
      markup = g_strconcat ("<small>", basename_utf8, "</small>", NULL);
      g_free (basename_utf8);
    }
  else
    {
      markup = g_strconcat ("<small>", _(DEFAULT_LABEL), "</small>", NULL);
    }

  gtk_label_set_markup (GTK_LABEL (button->_priv->label), markup);
}


/* ******************************* *
 *  GtkFileChooser Implementation  *
 * ******************************* */

static gboolean
esco_image_button_select_path (GtkFileChooser *chooser, const GtkFilePath *path, GError **error)
{
  EscoImageButton *button;
  GtkFileSystem *fs;
  gchar *filename;
  gboolean retval;

  button = ESCO_IMAGE_BUTTON (chooser);

  retval = _gtk_file_chooser_select_path (GTK_FILE_CHOOSER (button->_priv->dialog), path, error);

  if (retval)
    {
      fs = _gtk_file_chooser_get_file_system (GTK_FILE_CHOOSER (button->_priv->dialog));
      filename = gtk_file_system_path_to_filename (fs, path);

      if (filename != NULL)
	{
	  retval = try_set_button_image (button, filename, error);
	  g_free (filename);

	  if (retval)
	    reset_button_label (button);
	}
  }

  return retval;
}


static void
esco_image_button_file_chooser_iface_init (GtkFileChooserIface *iface)
{
  _gtk_file_chooser_delegate_iface_init (iface);

  iface->select_path = esco_image_button_select_path;
}


/* ******************* *
 *  GType Declaration  *
 * ******************* */

G_DEFINE_TYPE_WITH_CODE (EscoImageButton, esco_image_button, GTK_TYPE_TOGGLE_BUTTON, { \
  G_IMPLEMENT_INTERFACE (GTK_TYPE_FILE_CHOOSER, esco_image_button_file_chooser_iface_init) \
});


/* ********************** *
 *  Sub-Widget Callbacks  *
 * ********************** */

static void
dialog_update_preview_cb (GtkFileChooser * dialog, EscoImageButton * button)
{
  gchar *filename;
  gboolean active;

  filename = gtk_file_chooser_get_filename (dialog);
  active = FALSE;

  if (filename != NULL)
    {
      GdkPixbuf *pixbuf;
      gint width, height;
      gchar *type;

      width = 0;
      height = 0;

      pixbuf = util_load_pixbuf_for_size (filename, 192, 192, &width, &height, &type, NULL);

      if (pixbuf != NULL)
	{
	  gchar *markup;

	  if (type != NULL)
	    {
	      if (width < 0 || height < 0)
		markup = g_strconcat ("<small>", type, "\n", _("Any Size"), "</small>", NULL);
	      else
		markup = g_strdup_printf ("<small>%s\n%dx%d</small>", type, width, height);
	    }
	  else
	    {
	      if (width < 0 || height < 0)
		markup = g_strconcat ("<small>", _("Any Size"), "</small>", NULL);
	      else
		markup = g_strdup_printf ("<small>%dx%d</small>", width, height);
	    }

	  gtk_label_set_markup (GTK_LABEL (button->_priv->preview_label), markup);
	  g_free (markup);

	  gtk_image_set_from_pixbuf (GTK_IMAGE (button->_priv->preview_image), pixbuf);
	  g_object_unref (pixbuf);
	  active = TRUE;
	}

      g_free (type);
      g_free (filename);
    }

  gtk_file_chooser_set_preview_widget_active (dialog, active);
}


static void
dialog_response_cb (GtkFileChooser * dialog, gint response, EscoImageButton * button)
{
  gboolean hide_dialog;

  if (response == GTK_RESPONSE_ACCEPT)
    {
      if (gtk_file_chooser_get_preview_widget_active (dialog))
	{
	  hide_dialog = TRUE;

	  if (button->_priv->update_id != 0)
	    g_source_remove (button->_priv->update_id);

	  reset_button_image (button);
	  reset_button_label (button);

	  g_signal_emit_by_name (button, "selection-changed");
	}
      else
	{
	  hide_dialog = FALSE;
	}
    }
  else
    {
      hide_dialog = TRUE;
    }

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), !hide_dialog);
}


static void
dialog_current_folder_changed_cb (GtkFileChooser *dialog, GtkFileChooser *button)
{
  g_signal_emit_by_name (button, "current-folder-changed");
}


static void
dialog_file_activated_cb (GtkFileChooser *dialog, GtkFileChooser *button)
{
  g_signal_emit_by_name (button, "file-activated");
}


static void
dialog_notify_cb (GtkFileChooser *dialog, GParamSpec *pspec, GObject *button)
{
  if (pspec->param_id >= GTK_FILE_CHOOSER_PROP_FIRST &&
      pspec->param_id <= GTK_FILE_CHOOSER_PROP_LAST)
    {
      g_object_notify (button, pspec->name);
    }
}


static void
image_size_allocate_cb (GtkImage *image, GtkAllocation *allocation, EscoImageButton *button)
{
  if (allocation->width == button->_priv->image_width
      && allocation->height == button->_priv->image_height)
    return;

  button->_priv->image_width = allocation->width;
  button->_priv->image_height = allocation->height;

  if (button->_priv->update_id == 0)
    {
      button->_priv->update_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) reset_button_image,
						  button, NULL);
    }
}


/* *************************** *
 *  GtkToggleButton Functions  *
 * *************************** */

static void
esco_image_button_toggled (GtkToggleButton * button)
{
  EscoImageButtonPrivate *priv = ESCO_IMAGE_BUTTON_GET_PRIVATE (button);

  if (gtk_toggle_button_get_active (button))
    {
      if (!GTK_WIDGET_VISIBLE (priv->dialog))
	{
	  GtkWidget *toplevel;

	  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (button));

	  if (GTK_WIDGET_TOPLEVEL (toplevel) && GTK_IS_WINDOW (toplevel))
	    {
	      if (GTK_WINDOW (toplevel) != gtk_window_get_transient_for (GTK_WINDOW (priv->dialog)))
		gtk_window_set_transient_for (GTK_WINDOW (priv->dialog), GTK_WINDOW (toplevel));

	      gtk_window_set_modal (GTK_WINDOW (priv->dialog),
				    gtk_window_get_modal (GTK_WINDOW (toplevel)));
	    }
	}

      gtk_window_present (GTK_WINDOW (priv->dialog));
    }
  else
    {
      gtk_widget_hide (priv->dialog);
    }

  if (GTK_TOGGLE_BUTTON_CLASS (esco_image_button_parent_class)->toggled != NULL)
    (*GTK_TOGGLE_BUTTON_CLASS (esco_image_button_parent_class)->toggled) (button);
}


/* ********************* *
 *  GtkObject Functions  *
 * ********************* */

static void
esco_image_button_destroy (GtkObject * object)
{
  EscoImageButtonPrivate *priv;

  priv = ESCO_IMAGE_BUTTON_GET_PRIVATE (object);

  if (priv->dialog != NULL)
    gtk_widget_destroy (priv->dialog);

  if (priv->update_id != 0)
    g_source_remove (priv->update_id);

  if (GTK_OBJECT_CLASS (esco_image_button_parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS (esco_image_button_parent_class)->destroy) (object);
}


/* ******************* *
 *  GObject Functions  *
 * ******************* */

static GObject *
esco_image_button_constructor (GType type, guint n_params, GObjectConstructParam * params)
{
  GObject *object;
  EscoImageButtonPrivate *priv;
  GtkWidget *vbox;
  GtkFileFilter *filter;
  GSList *formats;
  gchar *tmp;

  object = (*G_OBJECT_CLASS (esco_image_button_parent_class)->constructor) (type, n_params, params);

  priv = ESCO_IMAGE_BUTTON_GET_PRIVATE (object);

  priv->dialog = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
			       "file-system-backend", priv->filesystem, NULL);
  g_free (priv->filesystem);
  priv->filesystem = NULL;
  g_object_set_qdata (object, GTK_FILE_CHOOSER_DELEGATE_QUARK, priv->dialog);
  gtk_dialog_add_button (GTK_DIALOG (priv->dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT);
  gtk_dialog_add_button (GTK_DIALOG (priv->dialog), GTK_STOCK_OK, GTK_RESPONSE_ACCEPT);
  g_signal_connect (priv->dialog, "response", G_CALLBACK (dialog_response_cb), object);
  g_signal_connect (priv->dialog, "update-preview", G_CALLBACK (dialog_update_preview_cb), object);
  g_signal_connect (priv->dialog, "current-folder-changed",
		    G_CALLBACK (dialog_current_folder_changed_cb), object);
  g_signal_connect (priv->dialog, "file-activated", G_CALLBACK (dialog_file_activated_cb), object);
  g_signal_connect (priv->dialog, "notify", G_CALLBACK (dialog_notify_cb), object);
  /*
   * We don't connect to the "selection-changed" signal because we emit that from the response
   * callback. The "update-preview" signal is never proxied because we don't support preview
   * widgets.
   */
  g_object_add_weak_pointer (G_OBJECT (priv->dialog), (gpointer *) (&priv->dialog));

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All Image Files"));

  for (formats = gdk_pixbuf_get_formats (); formats != NULL;
       formats = g_slist_remove_link (formats, formats))
    {
      guint i;
      gchar **mime_types;

      mime_types = gdk_pixbuf_format_get_mime_types (formats->data);
      for (i = 0; mime_types[i] != NULL; i++)
	{
	  gtk_file_filter_add_mime_type (filter, mime_types[i]);
	  g_free (mime_types[i]);
	}

      g_free (mime_types);
    }
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (priv->dialog), filter);

  vbox = gtk_vbox_new (FALSE, 3);
  gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (priv->dialog), vbox);
  gtk_widget_show (vbox);

  priv->preview_image = gtk_image_new ();
  gtk_box_pack_start (GTK_BOX (vbox), priv->preview_image, FALSE, FALSE, 0);
  gtk_widget_show (priv->preview_image);

  priv->preview_label = gtk_label_new (NULL);
  gtk_label_set_use_markup (GTK_LABEL (priv->preview_label), TRUE);
  gtk_label_set_justify (GTK_LABEL (priv->preview_label), GTK_JUSTIFY_CENTER);
  gtk_box_pack_start (GTK_BOX (vbox), priv->preview_label, FALSE, FALSE, 0);
  gtk_widget_show (priv->preview_label);

  vbox = gtk_vbox_new (FALSE, 3);
  gtk_container_add (GTK_CONTAINER (object), vbox);
  gtk_widget_show (vbox);
  
  priv->image = gtk_image_new ();
  gtk_widget_set_size_request (priv->image, priv->min_size, priv->min_size);
  g_signal_connect (priv->image, "size-allocate", G_CALLBACK (image_size_allocate_cb), object);
  gtk_container_add (GTK_CONTAINER (vbox), priv->image);
  gtk_widget_show (priv->image);

  tmp = g_strconcat ("<small>", _(DEFAULT_LABEL), "</small>", NULL);
  priv->label = gtk_label_new (tmp);
  g_free (tmp);
  gtk_label_set_justify (GTK_LABEL (priv->label), GTK_JUSTIFY_CENTER);
  gtk_label_set_use_markup (GTK_LABEL (priv->label), TRUE);
  gtk_box_pack_start (GTK_BOX (vbox), priv->label, FALSE, FALSE, 0);

  return object;
}


static void
esco_image_button_set_property (GObject * object, guint id, const GValue * value,
				GParamSpec * pspec)
{
  EscoImageButton *button = ESCO_IMAGE_BUTTON (object);

  switch (id)
    {
    case PROP_MIN_SIZE:
      esco_image_button_set_minimum_size (button, g_value_get_uint (value));
      break;
    case PROP_SHOW_LABEL:
      if (g_value_get_boolean (value))
	gtk_widget_show (button->_priv->label);
      else
	gtk_widget_hide (button->_priv->label);
      break;

    case PROP_TITLE:
    case GTK_FILE_CHOOSER_PROP_FILTER:
    case GTK_FILE_CHOOSER_PROP_ACTION:
    case GTK_FILE_CHOOSER_PROP_LOCAL_ONLY:
    case GTK_FILE_CHOOSER_PROP_USE_PREVIEW_LABEL:
    case GTK_FILE_CHOOSER_PROP_EXTRA_WIDGET:
    case GTK_FILE_CHOOSER_PROP_SHOW_HIDDEN:
      g_object_set_property (G_OBJECT (button->_priv->dialog), pspec->name, value);
      break;

    case GTK_FILE_CHOOSER_PROP_FILE_SYSTEM_BACKEND:
      button->_priv->filesystem = g_value_dup_string (value);
      break;

    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET_ACTIVE:
    case GTK_FILE_CHOOSER_PROP_SELECT_MULTIPLE:
      g_warning ("%s:%s: GtkFileChoosers of type `%s` do not support modifying the `%s` property.",
		 G_STRFUNC, G_STRLOC, G_OBJECT_TYPE_NAME (object), pspec->name);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}


static void
esco_image_button_get_property (GObject * object, guint id, GValue * value, GParamSpec * pspec)
{
  EscoImageButtonPrivate *priv = ESCO_IMAGE_BUTTON_GET_PRIVATE (object);

  switch (id)
    {
    case PROP_MIN_SIZE:
      g_value_set_uint (value, priv->min_size);
      break;
    case PROP_SHOW_LABEL:
      g_value_set_boolean (value, GTK_WIDGET_VISIBLE (priv->label));
      break;

    case PROP_TITLE:
    case GTK_FILE_CHOOSER_PROP_ACTION:
    case GTK_FILE_CHOOSER_PROP_FILE_SYSTEM_BACKEND:
    case GTK_FILE_CHOOSER_PROP_FILTER:
    case GTK_FILE_CHOOSER_PROP_LOCAL_ONLY:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET:
    case GTK_FILE_CHOOSER_PROP_PREVIEW_WIDGET_ACTIVE:
    case GTK_FILE_CHOOSER_PROP_USE_PREVIEW_LABEL:
    case GTK_FILE_CHOOSER_PROP_EXTRA_WIDGET:
    case GTK_FILE_CHOOSER_PROP_SELECT_MULTIPLE:
    case GTK_FILE_CHOOSER_PROP_SHOW_HIDDEN:
      g_object_get_property (G_OBJECT (priv->dialog), pspec->name, value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}


/* ***************** *
 *  GType Functions  *
 * ***************** */

static void
esco_image_button_class_init (EscoImageButtonClass * class)
{
  GObjectClass *gobject_class;
  GtkObjectClass *gtkobject_class;
  GtkToggleButtonClass *togglebutton_class;
  GtkWidgetClass *widget_class;

  gobject_class = G_OBJECT_CLASS (class);
  gtkobject_class = GTK_OBJECT_CLASS (class);
  togglebutton_class = GTK_TOGGLE_BUTTON_CLASS (class);
  widget_class = GTK_WIDGET_CLASS (class);

  gobject_class->constructor = esco_image_button_constructor;
  gobject_class->set_property = esco_image_button_set_property;
  gobject_class->get_property = esco_image_button_get_property;

  gtkobject_class->destroy = esco_image_button_destroy;

  togglebutton_class->toggled = esco_image_button_toggled;

  g_object_class_install_property (gobject_class, PROP_TITLE,
				   g_param_spec_string ("title", "Title",
							"The title of the file chooser dialog.",
							_(DEFAULT_TITLE), G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MIN_SIZE,
				   g_param_spec_uint ("minimum-size", "Minimum Size",
						      "The minimum size of the image.",
						      0, G_MAXUINT, DEFAULT_MIN_SIZE,
						      G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SHOW_LABEL,
				   g_param_spec_boolean ("show-label", "Show Label",
							 "Whether or not to show the image "
							 "filename label in the button.", TRUE,
							 G_PARAM_READWRITE));

  _gtk_file_chooser_install_properties (gobject_class);

  g_type_class_add_private (class, sizeof (EscoImageButtonPrivate));
}


static void
esco_image_button_init (EscoImageButton * button)
{
  button->_priv = G_TYPE_INSTANCE_GET_PRIVATE (button, ESCO_TYPE_IMAGE_BUTTON,
					       EscoImageButtonPrivate);

  button->_priv->min_size = DEFAULT_MIN_SIZE;

  button->_priv->src_width = -1;
  button->_priv->src_height = -1;
}


/* ************************************************************************** *
 *  Public API                                                                *
 * ************************************************************************** */

/**
 * esco_image_button_new:
 * @title: the title of the browse dialog.
 * 
 * Creates a new file-selecting entry widget.
 * 
 * Returns: a new entry widget.
 * 
 * Since: 2.6
 **/
GtkWidget *
esco_image_button_new (const gchar * title)
{
  return g_object_new (ESCO_TYPE_IMAGE_BUTTON, "title", title, NULL);
}


/**
 * esco_image_button_set_title:
 * @entry: the entry widget to modify.
 * @title: the new browse dialog title.
 * 
 * Modifies the @title of the browse dialog used by @entry.
 * 
 * Since: 2.6
 **/
void
esco_image_button_set_title (EscoImageButton * entry, const gchar * title)
{
  g_return_if_fail (ESCO_IS_IMAGE_BUTTON (entry));

  gtk_window_set_title (GTK_WINDOW (entry->_priv->dialog), title);
  g_object_notify (G_OBJECT (entry), "title");
}


/**
 * esco_image_button_get_title:
 * @entry: the entry widget to examine.
 * 
 * Retrieves the title of the browse dialog used by @entry. The returned value
 * should not be modified or freed.
 * 
 * Returns: a pointer to the browse dialog's title.
 * 
 * Since: 2.6
 **/
G_CONST_RETURN gchar *
esco_image_button_get_title (EscoImageButton * entry)
{
  g_return_val_if_fail (ESCO_IS_IMAGE_BUTTON (entry), NULL);

  return gtk_window_get_title (GTK_WINDOW (entry->_priv->dialog));
}


/**
 * esco_image_button_set_minimum_size:
 * @button: the widget to modify.
 * @min_size: the new minimum size of the image to use.
 * 
 * Changes the minimum size of the image preview in @button to @min_size.
 * 
 * <note>The minimum size is just that, an absolute minimum width/height that
 * the image preview inside @button will use. If @button is packed into a
 * container which allows size increases, both the button and the preview will
 * expand to fill the available space.</note>
 * 
 * Since: 2.6
 **/
void
esco_image_button_set_minimum_size (EscoImageButton * button, guint min_size)
{
  g_return_if_fail (ESCO_IS_IMAGE_BUTTON (button));

  button->_priv->min_size = min_size;
  gtk_widget_set_size_request (button->_priv->image, min_size, min_size);

  g_object_notify (G_OBJECT (button), "minimum-size");
}


/**
 * esco_image_button_get_minimum_size:
 * @button: the widget to examine.
 * 
 * Retrieves the minimum size of the @button image.
 * 
 * Returns: an unsigned integer representing the contents of the "minimum-size" property.
 * 
 * Since: 2.6
 **/
guint
esco_image_button_get_minimum_size (EscoImageButton * button)
{
  g_return_val_if_fail (ESCO_IS_IMAGE_BUTTON (button), 0);

  return button->_priv->min_size;
}


/**
 * esco_image_button_set_show_label:
 * @button: the widget to modify.
 * @show: whether or not to show the button label.
 * 
 * Sets the "show-label" property of @button to @show.
 * 
 * Since: 2.6
 **/
void
esco_image_button_set_show_label (EscoImageButton * button, gboolean show)
{
  g_return_if_fail (ESCO_IS_IMAGE_BUTTON (button));

  if (show)
    gtk_widget_show (button->_priv->label);
  else
    gtk_widget_hide (button->_priv->label);

  g_object_notify (G_OBJECT (button), "show-label");
}


/**
 * esco_image_button_get_show_label:
 * @button: the widget to examine.
 * 
 * Retrieves whether or not the label inside of @button is visible.
 * 
 * Returns: a boolean representing the contents of the "show-label" property.
 * 
 * Since: 2.6
 **/
gboolean
esco_image_button_get_show_label (EscoImageButton * button)
{
  g_return_val_if_fail (ESCO_IS_IMAGE_BUTTON (button), FALSE);

  return GTK_WIDGET_VISIBLE (button->_priv->label);
}
